"""
Methods to access:
    ls-tree
"""
import re
import subprocess
from pathlib import Path
from typing import Optional

from .constants import LS_TREE_LONG_RE, LS_TREE_RE, NOT_VALID_OBJECT_NAME_RE
from .datatypes import TreeContent
from .exceptions import GitException, UnknownRevisionException

__all__ = ["ls_tree"]


def __ls_tree_process_line(line: str) -> TreeContent:
    match = re.match(LS_TREE_RE, line)
    if match is None:
        raise ValueError("regex must have match")
    groups = match.groups()

    return TreeContent.from_str_values(
        mode=groups[0],
        type_=groups[1],
        object_=groups[2],
        file=groups[3],
    )


def __ls_tree_process_line_long(line: str) -> TreeContent:
    match = re.match(LS_TREE_LONG_RE, line)
    if match is None:
        raise ValueError("regex must have match")
    groups = match.groups()

    return TreeContent.from_str_values(
        mode=groups[0],
        type_=groups[1],
        object_=groups[2],
        object_size=groups[3],
        file=groups[4],
    )


def ls_tree(
        git_repo: Path,
        tree_ish: str,
        recursive: bool,
        use_long: bool,
        path: Optional[Path] = None) -> map:
    """
    Get the tree of objects in repo

        :param git_repo: Path to the repo
        :param tree_ish: The tree ish (branch name, HEAD)
        :param recursive: Whether tree is recursive
        :param use_long: Whether to get object sizes
        :param path: Filter path, defaults to None
        :raises UnknownRevisionException: Unknown tree_ish
        :raises GitException: Error to do with git
        :return: The git tree
    """
    args = ["git", "-C", str(git_repo), "ls-tree"]

    if use_long:
        args.append("-l")
    if recursive:
        args.append("-r")
    args.append(tree_ish)
    if path is not None:
        args.append(str(path))

    process_status = subprocess.run(args, capture_output=True)
    if not process_status.stdout and process_status.returncode != 0:
        stderr = process_status.stderr.decode()
        if re.match(NOT_VALID_OBJECT_NAME_RE, stderr):
            raise UnknownRevisionException(f"Unknown tree-ish '{tree_ish}'")
        raise GitException(stderr)

    split_lines = process_status.stdout.decode().strip().split("\n")

    if use_long:
        return map(__ls_tree_process_line_long, split_lines)
    return map(__ls_tree_process_line, split_lines)



"""
Methods for using git symbolic-ref command
"""
import subprocess
from pathlib import Path

from .exceptions import GitException, UnknownRefException

__all__ = [
    "change_symbolic_ref", "get_symbolic_ref",
    "delete_symbolic_ref", "get_active_branch",
    "change_active_branch",
]


def _raise_known_errors(process_status: subprocess.CompletedProcess, ref: str):
    """
    Used to raise any known git
    exceptions for the 'symbolic-ref' command

        :param process_status: The current procress status
        :param ref: The given reference/name
        :raises UnknownRefException: Unknown reference given
        :raises GitException: Error to do with git
    """
    if process_status.returncode == 1:
        raise UnknownRefException(f"Unknown ref {ref}")
    elif process_status.returncode > 0:
        raise GitException(process_status.stderr.decode())


def change_symbolic_ref(git_repo: Path, name: str, ref: str):
    """
    Change a symbolic ref in repo

        :param git_repo: Path to the repo
        :param name: The name (for example HEAD)
        :param ref: The reference
        :raises UnknownRefException: Unknown reference given
        :raises GitException: Error to do with git
    """
    args = ["git", "-C", str(git_repo), "symbolic-ref", name, ref]
    process_status = subprocess.run(args, capture_output=True)
    _raise_known_errors(process_status, ref)


def get_symbolic_ref(git_repo: Path, name: str) -> str:
    """
    Get a symbolic ref in repo

        :param git_repo: Path to the repo
        :param name: The name (for example HEAD)
        :raises UnknownRefException: Unknown reference given
        :raises GitException: Error to do with git
    """
    args = ["git", "-C", str(git_repo), "symbolic-ref", name]
    process_status = subprocess.run(args, capture_output=True)
    _raise_known_errors(process_status, name)
    return process_status.stdout.decode().strip()


def delete_symbolic_ref(git_repo: Path, name: str):
    """
    Delete a symbolic ref in repo

        :param git_repo: Path to the repo
        :param name: The name (for example HEAD)
        :raises UnknownRefException: Unknown reference given
        :raises GitException: Error to do with git
    """
    args = ["git", "-C", str(git_repo), "symbolic-ref", "-d", name]
    process_status = subprocess.run(args, capture_output=True)
    _raise_known_errors(process_status, name)


def get_active_branch(git_repo: Path) -> str:
    """
    Get the active (HEAD) reference

        :param git_repo: Path to the repo
        :raises UnknownRefException: Unknown reference given
        :raises GitException: Error to do with git
    """
    return get_symbolic_ref(git_repo, "HEAD")


def change_active_branch(git_repo: Path, branch: str):
    """
    Change the active (HEAD) reference

        :param git_repo: Path to the repo
        :param branch: The branch name to use
        :raises UnknownRefException: Unknown reference given
        :raises GitException: Error to do with git
    """
    change_symbolic_ref(git_repo, "HEAD", f"refs/heads/{branch}")



"""
Custom types that are used
"""
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Optional

__all__ = ["Log", "ArchiveTypes"]


class ArchiveTypes(Enum):
    """
    Possible archive types
    """
    TAR = "tar"
    TAR_GZ = "tar.gz"
    ZIP = "zip"


class TreeContentTypes(Enum):
    TREE = "tree"
    BLOB = "blob"


@dataclass
class Log:
    """
    Represents a single git log
    """
    commit_hash: str
    author_email: str
    commit_date: datetime
    subject: str


@dataclass
class TreeContent:
    """
    Reperesents a single 'ls-tree' entry
    """
    mode: str
    type_: TreeContentTypes
    object_: str
    file: Path
    object_size: Optional[int] = None

    @classmethod
    def from_str_values(cls, **kwargs):
        if not isinstance(kwargs["type_"], TreeContentTypes):
            kwargs["type_"] = TreeContentTypes(kwargs["type_"])
        if kwargs.get("object_size"):
            if kwargs["object_size"] == "-":
                kwargs["object_size"] = None
            else:
                kwargs["object_size"] = int(kwargs["object_size"])
        kwargs["file"] = Path(kwargs["file"])
        return cls(**kwargs)



"""
Constants
"""
EMPTY_REPO_RE = r"fatal: your current branch '\w+' does not have any commits yet"
UNKNOWN_REV_RE = r"fatal: ambiguous argument '\w+': unknown revision or path not in the working tree."
BRANCH_ALREADY_EXISTS_RE = r"fatal: A branch named '\w+' already exists."
BRANCH_NOT_FOUND_RE = r"error: branch '\w+' not found."
BRANCH_REFNAME_NOT_FOUND_RE = r"error: refname \w+ not found"
NOT_VALID_OBJECT_NAME_RE = r"fatal: Not a valid object name \w+"
# git ls-tree <tree-ish>
LS_TREE_RE = r"^(\d{6}) (\w+) (\w{40})\t(.+)$"
# git ls-tree <tree-ish> -l
LS_TREE_LONG_RE = r"^(\d{6}) (\w+) (\w{40}) +(\d+|\-)\t(.+)$"



"""
Methods for using the git branch command
"""
import re
import subprocess
from pathlib import Path
from typing import Tuple

from .constants import (BRANCH_ALREADY_EXISTS_RE, BRANCH_NOT_FOUND_RE,
                        BRANCH_REFNAME_NOT_FOUND_RE)
from .exceptions import (AlreadyExistsException, GitException,
                         NoBranchesException)

__all__ = [
    "get_branches", "new_branch",
    "copy_branch", "rename_branch",
    "delete_branch",
]


def get_branches(git_repo: Path) -> Tuple[str, Tuple[str]]:
    """
    Get the head branch and all others

        :param git_repo: Path to the repo
        :raises GitException: Error to do with git
        :raises NoBranchesException: Repo has no branches
        :return: the head branch and other branches
    """
    head = ""
    other_branches = []

    args = ["git", "-C", str(git_repo), "branch", "--no-color"]

    process_status = subprocess.run(args, capture_output=True)
    if not process_status.stdout:
        stderr = process_status.stderr.decode()
        if process_status.returncode != 0:
            raise GitException(stderr)
        if not stderr:
            raise NoBranchesException(
                f"no branches found for '{git_repo.name}'")

    split = process_status.stdout.decode().strip().split("\n")

    for line in split:
        if line[0] == "*":
            head = line[2:]
        else:
            other_branches.append(line[2:])

    return head, tuple(other_branches)


def new_branch(git_repo: Path, branch_name: str):
    """
    Create a new branch in repo

        :param git_repo: Path to the repo
        :param branch_name: Branch name
        :raises AlreadyExistsException: Branch already exists
        :raises GitException: Error to do with git
    """
    args = ["git", "-C", str(git_repo), "branch", branch_name]
    process_status = subprocess.run(args, capture_output=True)
    if process_status.returncode != 0:
        stderr = process_status.stderr.decode()
        if re.match(BRANCH_ALREADY_EXISTS_RE, stderr):
            raise AlreadyExistsException(
                f"branch name '{branch_name}' already exists")
        raise GitException(stderr)


def copy_branch(git_repo: Path, branch_name: str, new_branch: str):
    """
    Copy an existing branch to a new branch in repo (uses --force)

        :param git_repo: Path to the repo
        :param branch_name: Branch name
        :param new_branch: The new branch name
        :raises NoBranchesException: Branch does not exist
        :raises GitException: Error to do with git
    """
    args = ["git", "-C", str(git_repo), "branch", "-C",
            branch_name, new_branch]
    process_status = subprocess.run(args, capture_output=True)
    if process_status.returncode != 0:
        stderr = process_status.stderr.decode()
        if re.match(BRANCH_REFNAME_NOT_FOUND_RE, stderr):
            raise NoBranchesException(
                f"no branch found with name '{branch_name}'")
        raise GitException(stderr)


def rename_branch(git_repo: Path, branch_name: str, new_branch: str):
    """
    Rename an existing branch (uses --force)

        :param git_repo: Path to the repo
        :param branch_name: Branch name
        :param new_branch: The new branch name
        :raises NoBranchesException: Branch does not exist
        :raises GitException: Error to do with git
    """
    args = ["git", "-C", str(git_repo), "branch", "-M",
            branch_name, new_branch]
    process_status = subprocess.run(args, capture_output=True)
    if process_status.returncode != 0:
        stderr = process_status.stderr.decode()
        if re.match(BRANCH_REFNAME_NOT_FOUND_RE, stderr):
            raise NoBranchesException(
                f"no branch found with name '{branch_name}'")
        raise GitException(stderr)


def delete_branch(git_repo: Path, branch_name: str):
    """
    Delete an existing branch (uses --force)

        :param git_repo: Path to the repo
        :param branch_name: Branch name
        :raises NoBranchesException: Branch does not exist
        :raises GitException: Error to do with git
    """
    args = ["git", "-C", str(git_repo), "branch", "-D", branch_name]
    process_status = subprocess.run(args, capture_output=True)
    if process_status.returncode != 0:
        stderr = process_status.stderr.decode()
        if re.match(BRANCH_NOT_FOUND_RE, stderr):
            raise NoBranchesException(
                f"no branch found with name '{branch_name}'")
        raise GitException(stderr)
